"""
..
Copyright 2020 Twinleaf LLC
License: MIT
"""

from typing import List
import tio
import struct

def rpc_capture_block(self, index:int = 0, blocksize:int=256, size:int=32768, typecode:str="L") -> List[float]:
  capture = []
  blocks = int(size/blocksize)
  entriesPerBlock = int(blocksize/struct.Struct('<'+typecode).size)
  packcode = "<"+str(entriesPerBlock)+typecode
  for blockindex in range(blocks):
    block = self._tio.rpc_val('capture.block', rpcType = tio.UINT16_T, value=blockindex, returnRaw = True)
    capture += struct.unpack(packcode, block)
  return capture





#!/usr/bin/env python3
"""
..
    Copyright: 2018 Twinleaf LLC
    Author: kornack@twinleaf.com

Print stream and source information.

"""

import tldevice

def rateReport(dev):
  timebase_id = dev._tio.protocol.streamInfo['stream_timebase_id']
  print(f"Stream 0: timebase ID: {timebase_id}"
       +f", components: {dev._tio.protocol.streamInfo['stream_total_components']}"
       +f", period: {dev._tio.protocol.streamInfo['stream_period']}")
  print(f"Timebase {timebase_id} rate: {dev._tio.protocol.timebases[timebase_id]['timebase_Fs']:.3f} Hz"
       +f" ({dev._tio.protocol.timebases[timebase_id]['timebase_period_num_us']}/{dev._tio.protocol.timebases[timebase_id]['timebase_period_denom_us']} µs)"
       +f", epoch: {dev._tio.protocol.timebases[timebase_id]['timebase_epoch']}"
       +f", stability: {dev._tio.protocol.timebases[timebase_id]['timebase_stability_ppb']:.0f} ppb")
  for i, column in enumerate(dev._tio.protocol.streams):
    print(f"Component {i}: {column['source_name']} ({column['source_title']}), period {column['source_period']}, {column['stream_Fs']:.3f} Hz")

if __name__ == "__main__":
  import argparse
  parser = argparse.ArgumentParser(prog='tio_rates', 
                                   description='Live Twinleaf I/O Data Stream Monitor.')
  parser.add_argument("url", 
                      nargs='?', 
                      default='tcp://localhost/',
                      help='URL: tcp://localhost')
  args = parser.parse_args()
  device = tldevice.Device(url=args.url)
  rateReport(device)




"""
..
    Copyright: 2017-2021 Twinleaf LLC
    Author: kornack@twinleaf.com
    OriginalDate: November 2017
"""

import tio

class TwinleafDataController(object):
  def __init__(self, dev):
    self._dev = dev

  def __call__(self, samples=1, duration=None, timeaxis=False, flush=True):
    return self._dev._tio.stream_read_raw(samples = samples, duration=duration, flush=flush, timeaxis=timeaxis)

  def columnnames(self, withName=True):
    columnnames = self._dev._tio.protocol.columns
    if withName:
      routingString = "/"+"/".join(map(str,self._dev._tio.routing))
      columnnames = [self._dev._tio.name+' '+routingString+' '+columnname for columnname in columnnames ]
    return columnnames

  def iter(self, samples=0, flush=True, timeaxis=False):
    if flush:
      self._dev._tio.pub_flush()
    if samples==0:
      while True:
        self._dev._tio.pub_warn_overload()
        yield self._dev._tio.stream_read_raw(samples = 1, flush=False, timeaxis=timeaxis)
    else:
      for x in range(samples):
        self._dev._tio.pub_warn_overload()
        yield self._dev._tio.stream_read_raw(samples = 1, flush=False, timeaxis=timeaxis)

  def queueSize(self):
    return self._dev._tio.pub_queue.qsize()

  def rate(self, value=None):
    if value is None:
      return self._dev._tio.protocol.streams[0]['stream_Fs'] 
    else:
      return self._dev._tio.rpc_val('data.rate', rpcType = tio.FLOAT32_T, value=value)



#!/usr/bin/env python3
"""
..
    Copyright: 2017 Twinleaf LLC
    Author: kornack@twinleaf.com

Log data!

"""

import tldevice
import argparse

parser = argparse.ArgumentParser(prog='tio_log', 
                                 description='Very simple logging utility.')

parser.add_argument("url", 
                    nargs='?', 
                    default='tcp://localhost/',
                    help='URL: tcp://localhost')
parser.add_argument("logfile", 
                    nargs='?', 
                    default='log.tsv',
                    help='Log filename: log.tsv')
parser.add_argument("--rpc", 
                    action='append', 
                    default=[],
                    type=lambda kv: kv.split(":"), 
                    help='Commands to be run on start; rpc:type:val')
args = parser.parse_args()

device = tldevice.Device(url=args.url, rpcs=args.rpc)

file = open(args.logfile,'w') 

print(f"Logging to {args.logfile} ...")

file.write("\t".join(map(str,device.data.columnnames()))+"\n")

for row in device.data.iter():
  # Tab delimited data
  rowstring = "\t".join(map(str,row))+"\n"
  # Write line
  file.write(rowstring)





#!/usr/bin/env python3
"""
itio: Interactive Twinleaf I/O 
License: MIT
Author: Thomas Kornack <kornack@twinleaf.com>
"""

import tldevicesync
import argparse

def main():
  parser = argparse.ArgumentParser(prog='itio', 
                                   description='Interactive Twinleaf I/O.')
  
  parser.add_argument("url", 
                      nargs='?', 
                      default='tcp://localhost/',
                      help='URL: tcp://localhost')
  parser.add_argument("--rpc", 
                      action='append', 
                      default=[],
                      type=lambda kv: kv.split(":"), 
                      help='Commands to be sent on start; rpc:type:val')
  parser.add_argument('-v', 
                      action="store_true",
                      default=False,
                      help='Verbose output for debugging')
  parser.add_argument('-r', 
                      action="store_false",
                      default=True,
                      help='Ignore and rebuild rpc/stream cache')
  parser.add_argument('-t', 
                      action="store_true",
                      default=False,
                      help='Quit after 1 second network socket timeout')
  args = parser.parse_args()
  
  device = tldevicesync.DeviceSync(url=args.url, verbose=args.v, rpcs=args.rpc, stateCache=args.r, timeout=args.t)
  device._interact()

if __name__ == "__main__":
  main()



#!/usr/bin/env python3
# coding: utf-8
"""
Twinleaf Generic Multi Device Control
Copyright 2019 Twinleaf LLC
License: MIT
"""

import tldevice
import threading
import time

class DeviceSync():
  def __init__(self, url="tcp://localhost", verbose=False, rpcs=[], stateCache=True, connectingMessage = True, connectionTime = 1, timeout = False):
    self._routes = {}
    self._routes["/"] = tldevice.Device(url=url, verbose=verbose, rpcs=rpcs, stateCache=stateCache, connectingMessage=connectingMessage, timeout = timeout)
    self._routes["/"]._tio.recv_router = self._recvRouter
    self.__dict__[self._routes["/"]._shortname] = self._routes["/"]
    time.sleep(connectionTime)

  def _recvRouter(self, routing, packet):
    routingKey = '/'.join(map(str,routing))
    if routingKey in self._routes.keys():
        self._routes[routingKey]._tio.recv_queue.put(packet)
    else: # Create new route
      #print(f"Creating route to {routingKey}.")
      self._routes[routingKey] = tldevice.Device(url="router://interthread/"+routingKey, send_router = self._routes["/"]._tio.send, verbose=True, specialize=False)
      threading.Thread(target=self._specialize, args=(routingKey,)).start()

  def _specialize(self, routingKey):
    self._routes[routingKey]._specialize()
    self._routes[routingKey]._shortname += routingKey
    #self._routes[routingKey]._tio.shortname = self._routes[routingKey]._shortname
    self.__dict__[self._routes[routingKey]._shortname.replace("/","")] = self._routes[routingKey]

  def _interact(self):
    imported_objects = {}
    imported_objects['tio'] = self
    banner=""
    exit_msg = f"tio thanks you."
    try:
      from IPython import embed
      embed(
        user_ns=imported_objects, 
        banner1=banner, 
        banner2=f"Use   : tio.<tab>",
        exit_msg=exit_msg)
    except ImportError:
      import code
      repl = code.InteractiveConsole(locals=imported_objects)
      repl.interact(
        banner=banner, 
        exitmsg = exit_msg)

class SyncStream():
  def __init__(self, streams = []):
    self.streams = streams
    self.sync()

  def sync(self, flush=True):
    # Find the initial datum time
    times = []
    data = []
    for stream in self.streams:
      row = stream(samples=1, flush=flush, timeaxis=True)
      times += [row[0]]
      data += row[1:]

    # TODO: Check that the streams have compatible data rates

    # Ensure the times match up!
    # If not, catch up on the streams that are behind
    maxtime = max(times)
    mintime = min(times)
    if maxtime != mintime:
      for i,stream in enumerate(self.streams):
        max_deviation = 0
        while times[i] < maxtime:
          # print(f"Drop a sample on stream {i}")
          times[i] = stream(samples=1, flush=False, timeaxis=True)[0]
          max_deviation -= 1
          if max_deviation > 5:
            raise Exception("Can't sync stream!")

  def __call__(self, samples = 1, duration=None, timeaxis=True, flush=True):
    return self.read(samples = samples, duration=duration, timeaxis=timeaxis, flush=flush)

  def read(self, samples = 1, duration=None, timeaxis=True, flush=True):
    if flush:
      self.sync()

    # Acquire data
    times = []
    data = [] 
    for stream in self.streams:
      streamdata = stream(samples=samples, duration=duration, timeaxis=True, flush=False)
      times += [streamdata[0]]
      data += streamdata[1:]
    
    if samples == 1:
      starttimes = times
    else:
      starttimes = [timecol[0] for timecol in times]

    if max(starttimes) != min(starttimes):
      delta = max(starttimes) - min(starttimes)
      raise Exception(f"Streams out of sync by {delta}!")
    
    if timeaxis:
      data = [times[0]] + data

    return data

  def readQueueSize(self):
    """This reports the queue depth for the first stream"""
    return self.streams[0].queueSize()

  def readAvailable(self, timeaxis=True):
    samples = self.readQueueSize()
    if samples < 1:
      samples = 1
    return self.read(samples=samples, timeaxis=timeaxis, flush=False)

  def columnnames(self, timeaxis=True, withName=True):
    if timeaxis:
      names = ["time"]
    else:
      names = []
    for stream in self.streams:
      names += stream.columnnames(withName=withName)
    return names

  def rate(self):
    return self.streams[0].rate()

  def iter(self, samples=0, flush=True):
    if flush:
      yield self.read(samples = 1, flush=True)
      samples -= 1
    if samples<=0:
      while True:
        yield self.read(samples = 1, flush=False)
    else:
      for x in range(number):
        yield self.read(samples = 1, flush=False)
  
if __name__ == "__main__":
  device = DeviceSync()
  device._interact()


#!/usr/bin/env python3
"""
..
    Copyright: 2018-2021 Twinleaf LLC
    Author: kornack@twinleaf.com

Parse native format logged data.

Assumes no SLIP encoding.

"""

import tio
import argparse
import hexdump
import logging
import struct
import time
import os

def main():
  
  parser = argparse.ArgumentParser(prog='tio_logfile', 
                                   description='Unpack raw TIO log files (.tio format).')
  
  parser.add_argument("logfile", 
                      type=str,
                      nargs='+', 
                      default='Log 000000.tio',
                      help='Filename like "Log 000000.tio"')
  parser.add_argument('-v', 
                      action="store_true",
                      default=False,
                      help='Verbose output for debugging')
  parser.add_argument('-vp', 
                      action="store_true",
                      default=False,
                      help='Extra verbose protocol output for debugging')
  parser.add_argument('--raw', 
                      action="store_true",
                      default=False,
                      help='Display output for each line')
  parser.add_argument('--lines', 
                      action="store",
                      default=None,
                      help='Limit number of packets to process')
  parser.add_argument('--sth', 
                      type=float,
                      default=10,
                      help='Factor by which slower streams are excluded from merge')
  parser.add_argument('--separate', 
                      action="store_true",
                      default=False,
                      help='Do not merge the parsed files.')
  
  args = parser.parse_args()
  
  logLevel = logging.ERROR
  if args.v:
    logLevel = logging.DEBUG
  logging.basicConfig(level=logLevel)
  logger = logging.getLogger('tio-logfile')
  
  # First read through and generate individual files for each stream.
  routes=[]
  sensors={}
  tempfilenames = {}
  tempfiles = {}
  firsttimes = {}
  datarates = {}
  lines = 0
  
  filenames = args.logfile
  for filename in filenames:
    if filename[-4:]==".tio":
      outputfile = filename[:-4]+".tsv"
    else:
      outputfile = filename+".tsv"
    with open(filename,'rb') as f:
      while True:
        lines += 1
        if args.lines is not None:
          if lines > int(args.lines):
            break
        header = bytes(f.read(4))
        if len(header) < 4:
          break
    
        headerFields = struct.unpack("<BBH", header )
        payloadType, routingSize, payloadSize = headerFields
        if payloadSize > tio.TL_PACKET_MAX_SIZE or routingSize > tio.TL_PACKET_MAX_ROUTING_SIZE:
          logger.error('Packet too big');
          raise
        else:
          payload = bytes(f.read(payloadSize+routingSize))
          packet = header+payload
          routingBytes = b''
          routingString = "/"
          if routingSize > 0:
            routingBytes = payload[-routingSize:]
            routingString += "/".join(map(str,list(routingBytes)))+"/"
            # If there's more than one file, prefix the routing name with the filename
          if len(filenames) > 1:
            routingString = os.path.basename(filename[:-4]) + routingString
          if routingString not in routes:
            routes += [routingString]
            sensors[routingString] = tio.TIOProtocol(verbose = args.vp, routing=list(routingBytes))
            tempfilenames[routingString] = outputfile[:-4]+f"-{list(routingBytes)}.tsv"
            tempfiles[routingString] = open(tempfilenames[routingString], 'w')
    
          try:
            parsedPacket = sensors[routingString].decode_packet(packet)
          except Exception as error:
            logger.debug('Error decoding packet:');
            hexdump.hexdump(packet)
            logger.exception(error)
          
          if args.raw:
            print(parsedPacket)
    
          if parsedPacket['type'] == tio.TL_PTYPE_STREAM0:
            row = sensors[routingString].stream_data(parsedPacket, timeaxis=True)
            if row !=[]:
              time,data = row
              if routingString not in firsttimes.keys():
                # This is the first row
                firsttimes[routingString] = time
                datarates[routingString] = sensors[routingString].streams[0]['stream_Fs'] 
                # Write out columns
                headerString = ""
                if len(sensors[routingString].columns)>0:
                  for column in ["time"]+sensors[routingString].columns:
                    headerString += routingString+column+"\t"
                  tempfiles[routingString].write(headerString[:-1]+"\n")
              rowsamples = len(data)
              rowstring = str(time)+"\t"
              rowstring += "\t".join(map(str,data))
              # Add blanks to pack out when absent data
              rowstring += "\t"*(len(sensors[routingString].columns)+1-rowsamples) # +1 for time column
              rowstring = rowstring[:-1]+"\n"
              tempfiles[routingString].write(rowstring)
  
  for fd in tempfiles.values():
    fd.close()
  
  print(f"Found data streams from routes:")
  [print(f"- {route}") for route in routes]
  
  if args.separate:
    exit
  
  # Remove routes that don't have valid start times
  for idx,route in enumerate(list(routes)): # copy the routes
    if route not in firsttimes.keys():
      print(f"NB: Not merging from route {route} because the timing metadata is missing.")
      tempfiles.pop(route)
      tempfilenames.pop(route)
      sensors.pop(route)
      routes.remove(route)

  
  try:
    firsttime = max(firsttimes.values())
  except:
    raise Exception("No metadata in sample enough to get a first sample. Sample for longer?")
  if firsttime>0:
    print(f"Writing (merged) log starting at {firsttime} s.")
  
  # If there are files with global timestamps, then set aside the files with local timestamps
  # heuristic for whether there are valid unix timestamps
  if firsttime>1000000000: # Sat Sep 08 2001 21:46:40 UTC-0400 (EDT)
    # Find all the files that use local time and remove them from merging
    for route, thisfirsttime in firsttimes.items():
      if thisfirsttime < 1000000000:
        print(f"NB: Not merging from route {routes[idx]} because its starting time {thisfirsttime} s does not appear to have a global timestamp.")
        tempfiles.pop(route)
        tempfilenames.pop(route)
        sensors.pop(route)
        # firsttimes.pop(route)
        datarates.pop(route)
        routes.remove(route)
      # else:
        # print(f"Found metadata for route {route} with starting time {thisfirsttime}.")
  
  # discardedTime = max(firsttimes.values()) - min(firsttimes.values())
  # if discardedTime > 0:
  #   print(f"NB: Discarding up to the first {discardedTime} seconds of data to merge the logs")
  
  # If there are streams with widely varying data rates, then set aside the streams with low rates
  slowerThreshold = args.sth
  dataratemax = max(datarates.values())
  for idx, datarate in enumerate(datarates.values()):
    if datarate < dataratemax / slowerThreshold:
      print(f"NB: Not merging from route {routes[idx]} because its data rate {datarate} Hz < dominant rate {dataratemax} Hz / {slowerThreshold}." )
      tempfiles.pop(routes[idx])
      tempfilenames.pop(routes[idx])
      sensors.pop(routes[idx])
      firsttimes.pop(routes[idx])
      routes.pop(idx)
  
  print(f"Merging data streams from routes:")
  [print(f"- {route} starting {firsttimes[route]}") for route in routes]
  
  # Now write out combined file
  tempfilelist = []
  for routingString in routes:
    fd = open(tempfilenames[routingString], 'r')
    tempfilelist += [ fd ]
  
  with open(outputfile,'w') as fout:
    headerString=""
    for fd in tempfilelist:
      headerString += fd.readline()[:-1] + "\t"
    headerString = headerString[:-1]+"\n"
    fout.write(headerString)
  
    while True:
      line = ""
      for fd in tempfilelist:
        linesegment = fd.readline()
        try:
          time = float(linesegment.split('\t')[0])
        except:
          break
        while time <= firsttime:
          linesegment = fd.readline()
          try:
            time = float(linesegment.split('\t')[0])
          except:
            raise IndexError("Could not find time alignment among streams")
        if linesegment != "":
          line += linesegment[:-1]+"\t"
      if line == "":
        break
      line = line[:-1]+"\n"
      fout.write(line)
  
  # close and delete temporary files
  for fd in tempfilelist:
    fd.close()
  for tempfile in tempfilenames.values():
    os.remove(tempfile)
  
if __name__ == "__main__":
  main()



#!/usr/bin/env python3
# coding: utf-8
# Until PEP 517 is more widespread, we use this stub to provide backward compatibility
# install using, for example, `python3 setup.py install`
import setuptools
setuptools.setup()
